/*

    File: compile.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#if !defined M_SCRIPT_COMPILE_H
#define M_SCRIPT_COMPILE_H

#include "interface.h"
#include "lexeme.h"
#include "program.h"

//test language

typedef struct {
	int		type;		//return type
	int		value;		//address
	int		offset;		//offset to address.  directly added to 'value' for internal vars.  specially-used for external vars
	bool	constant;	//const/addr		- overrides relative and write
	bool	write;		//read/write
	bool	relative;	//relative/global
} rw_addr_t;

class CCodeCompiler {
private:
	CLexemeParser *lex;
	CProgram *program;

	//build as you go
	//whistle while you work
	CFuncType currentFuncType;

	void error(const char *, ...);
	bool foundError;

#ifdef DATA_STRING
	string errorStr;
#else
	char errorStr[256];
#endif

	void typeError(const char *msg, int rightType, int wrongType);

	//<exprlist>	-> <expr> {"," <expr>}
	void exprlist(int *type, int size);
#ifdef ATTEMPT_IMPLEMENT_FORK
	void label_fork();	//builtin for ease of use =)
#endif
	bool label_var(bool write, rw_addr_t *rw);
	int complete_rw(rw_addr_t rw);
	int call_exprlist(int callTypeIndex);
	int label_process_call(rw_addr_t rw);
	//<label>	-> "fork" "{" <stmt> "}" | name ["(" [<exprlist>] ")"]
	int label();

	//<exprPar> -> "("<expr>")" | <label>
	int exprPar();
	//<exprUnary> -> ( - | ! | ~ ) <exprPar>
	int exprUnary();
	//<exprProd> -> <exprUnary> {( * | / | % ) <exprUnary>}
	int exprProd();
	//<exprSum> -> <exprProd> {( + | - ) <exprProd>}
	int exprSum();
	//<exprShift> -> <exprSum> {( << | >> ) <exprSum>}
	int exprShift();
	//<exprIneq> -> <exprShift> {( < | > | <= | >= ) <exprShift>}
	int exprIneq();
	//<exprEquiv> -> <exprIneq> {( == | != ) <exprIneq>}
	int exprEquiv();
	//<exprLogicAnd> -> <exprEquiv> {& <exprEquiv>}
	int exprLogicAnd();
	//<exprLogicXOr> -> <exprLogicAnd> {^ <exprLogicAnd>}
	int exprLogicXOr();
	//<exprLogicOr> -> <exprLogicXOr> {| <exprLogicXOr>}
	int exprLogicOr();
	//<exprAnd> -> <exprLogicOr> {&& <exprLogicOr>}
	int exprAnd();
	//<exprOr> -> <exprAnd> {|| <exprAnd>}
	int exprOr();

	//rules to add (in this location's order)
	//<exprCond> -> <exprOr> ? <exprOr> : <exprOr>
	int exprCond();

	//shortcut for whatever is our first expr()
	int expr() {return exprCond();}

	//<type>		-> typename ["(" [<type> {"," <type>} ] ")" ]
	int type(bool force);
	//<stmt>		-> ([<type>] varname_id ["=" <expr>] | <expr>) ";"
	void stmt_decl(int varType);
	void stmt_assign(int varIndex);
	void stmt_return();
	void stmt_if();
	void stmt_while();
	void stmt_for();
	void stmt();
	//<stmtblock>	-> "{" {<stmt>} "}"
	void stmtblock();
	//<param>		-> <type> paramname
	void param();
	//<paramlist>	-> "(" [<param> {"," <param>}] ")"
	void paramlist();
	//<func>		-> <type> funcname <paramlist> <stmtblock>
	void func();
	//<structure>	->	structname "{" {<type> <fieldName> ";"} "}"
	void structure();
	//<funcs>		-> {<func> | "struct" <structure>}
	void funcs();

public:
	CCodeCompiler(CLexemeParser *lex, CProgram *program);
	bool compile();
	const char *getError();
};

#endif